"""A simple timer context manager and decorator."""

import time
from collections.abc import Iterator
from contextlib import contextmanager


class Timer:
    """A basic code timer."""

    def __init__(self) -> None:
        """Create a timer."""
        self.start_time = time.perf_counter()
        self.end_time: float | None = None

    @property
    def elapsed(self) -> float:
        """Seconds elapsed since timer creation."""
        return time.perf_counter() - self.start_time


@contextmanager
def timer(print_elapsed: bool = False) -> Iterator[Timer]:
    """
    Time the indented block.
    Yields a Timer object with `elapsed` and `end_time` properties.
    """
    timer = Timer()
    yield timer
    timer.end_time = timer.elapsed
    if print_elapsed:
        print(f"{timer.end_time:.8f} seconds elapsed.")



"""Typed wrappers for functools.cache and functools.lru_cache decorators."""

import functools
from collections.abc import Callable
from typing import ParamSpec, TypeVar, cast

P = ParamSpec("P")
R = TypeVar("R")


def cache(func: Callable[P, R]) -> Callable[P, R]:
    """functools.cache wrapper with typed signature."""

    @functools.wraps(func)
    @functools.cache
    def _wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        return func(*args, **kwargs)

    return cast(Callable[P, R], _wrapper)


def lru_cache(
    maxsize: int = 128, typed: bool = False
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """functools.lru_cache wrapper with typed signature."""

    def _lru_cache(func: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(func)
        @functools.lru_cache(maxsize, typed)
        def _wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            return func(*args, **kwargs)

        return cast(Callable[P, R], _wrapper)

    return _lru_cache
